/* eslint-disable */
// tslint:disable
/**
 * FastAPI
 *
 *
 * OpenAPI spec version: 0.1.0
 *
 *
 * NOTE: This class is auto generated by OpenAPI Generator+.
 * https://github.com/karlvr/openapi-generator-plus
 * Do not edit the class manually.
 */

export * from "./api";
export * from "./models";
export * from "./configuration";
export { RequiredError } from "./runtime";
export type { FetchAPI, FetchArgs } from "./runtime";
